package PureMessage::Milter::Policy;

use strict;
use base 'PerlMx';

our $VERSION = "5.600";

use PerlMx;  # import SMFIS_* constants
use PureMessage::_sys ();
use Socket qw(unpack_sockaddr_in inet_ntoa);

use PureMessage::Config qw(expand_config_filename expand_filename %config);
use PureMessage::Policy::Engine;
use PureMessage::Policy::Events;
use PureMessage::Policy::MessageObject;
use PureMessage::Policy::Logger;
use PureMessage::Quarantine;
use PureMessage::MessageInfo;
use Mail::Field;

require ActiveState::StopWatch if $config{time_connection} || $config{time_message};

# Some modules that are used indirectly but would otherwise end
# up loaded by demand.  We load them here to make code sharing between
# milter children more effective.
require MIME::Base64;
require MIME::QuotedPrint;
require Carp::Heavy;

$| = 1;

# log priority constants (borrowed from syslog)
use PerlMx::Log qw(LOG_DEBUG LOG_INFO LOG_NOTICE LOG_WARNING
                   LOG_ERR LOG_CRIT LOG_ALERT LOG_EMERG);

$SIG{__WARN__} = sub { __PACKAGE__->log(LOG_WARNING, @_) };


# milter processing callbacks
my %callbacks = (
		  CONNECT => \&pmx_connect,
		  HELO    => \&pmx_helo,
		  ENVFROM => \&pmx_envfrom,
		  ENVRCPT => \&pmx_envrcpt,
		  EOH     => \&pmx_eoh,
		  BODY    => \&pmx_body,
		  EOM     => \&pmx_eom,
		  ABORT   => \&pmx_abort,
		  CLOSE   => \&pmx_close,
		);

my $hostname = PureMessage::_sys::gethostname();
my $quarantine = PureMessage::Quarantine->new(undef, logger => __PACKAGE__);
$quarantine->new_tempfile;  # make sure tempfile class is loaded early

my $splitting;
my $engine; _MakeEngine();

my $block_size = $config{block_size} || 8192;
my $relay_header = $config{relay_from_header};
my $relay_index = $config{relay_from_index};
my $relay_use_header = $config{relay_use_header};
my $relay_regex = qr/$config{relay_from_regex}/;
my $def_relay_regex = PureMessage::Config->new->default("relay_from_regex") || '';
my $UNKNOWN_RELAY_IP = '0.0.0.0';

if (__PACKAGE__->smtp_mode()){
    $relay_use_header = 1;
}
sub _MakeEngine {
    my $script = shift || expand_filename("etc/policy.siv");
    $engine = PureMessage::Policy::Engine->new(
        script => $script,
        plugins => [expand_filename("etc/data/policy/base.conf")],
        quarantine => $quarantine,
        quiet => 1,
        logger => PureMessage::Policy::Logger->new(logger => __PACKAGE__,
						   tee_stdout => 1),
        init => 1,
        message_exception_action => $config{message_exception_action},        
    );

    __PACKAGE__->log(LOG_INFO, __PACKAGE__ . " up running with " .
		               (join(' ', $engine->loaded_plugins) ||
				"no plugins")
                               . '.');
    $engine->logger->tee_stdout(0);
    $splitting = $engine->one_recipient_per_message || !$config{avoid_message_splitting};
    return $engine;
}

sub new {
    my $class = shift;
    my $self = bless { FLAGS => SMFIF_ADDHDRS | SMFIF_CHGHDRS |
			        SMFIF_CHGBODY | SMFIF_DELRCPT |
                                SMFIF_ADDRCPT,
                # | SMFIF_CHGFROM, # XXX: Add back in to support changing env from
		       %callbacks,
		     }, $class;
    $self;
}

sub pmx_connect {
    my ($self, $peer_hostname, $peer_sockaddr) = @_;
    
    $self->{mcount} = 0;
    $self->{watch} = ActiveState::StopWatch::start_watch()
	if $config{time_connection};
    
    my($port, $a_addr);
    if (defined $peer_sockaddr and length $peer_sockaddr) {
        my $n_addr;
	($port, $n_addr) = unpack_sockaddr_in($peer_sockaddr);
	$a_addr = inet_ntoa($n_addr);
    }
    else {
	# this typically means peer is another sendmail on the local
	# box running in local delivery mode (so there is no peer)
	$port = 0;
	$a_addr = "0.0.0.0";
    }
    $self->{con}{'Relay'} = $peer_hostname;
    $self->{con}{'Relay'} .= (index($peer_hostname,'[') == 0) ? '' : " [$a_addr]";

    my $sm_host = $self->m_canonical_name;
    $self->{con}{'Sendmail-Host'} = $sm_host;
    $self->{con}{'Milter-Host'} = $hostname if $hostname ne $sm_host;

    # IP hash checks
    $self->{con}{'Relay-Info'} = $self->{__perlmx_relay_info}
	if exists $self->{__perlmx_relay_info};
        
    $self->log(LOG_DEBUG, "connect $peer_hostname $a_addr:$port\n");
    
    # create a dummy message object that will run against relay tests
    $self->{dummy} = PureMessage::Policy::MessageObject->new(
            minfo => PureMessage::MessageInfo->new(%{$self->{con}}),
            event => 'connect',
            logger => $self,
            quarantine => $quarantine,
    );
    
    # do not process if we are doing message splitting, or are in
    # SMTP mode.
    return SMFIS_CONTINUE if $splitting or $self->smtp_mode;
    
    return $self->_process($self->{dummy}, new=>1);
}

sub pmx_helo {
    my ($self, $host) = @_;

    $self->log(LOG_DEBUG, "helo: $host");
    $self->{con}{'Helo'} = $host;
    $self->{dummy}->minfo_value('Helo', $host) if exists $self->{dummy};

    return SMFIS_CONTINUE;
}

sub pmx_envfrom {
    my ($self, $sender, @esmtp_args) = @_;
    
    $self->log(LOG_DEBUG, "envfrom: $sender");

    $self->{msg_watch} = ActiveState::StopWatch::start_watch()
	if $config{time_message};

    # this is where a new message starts
    # if this is the first message, use the dummy
    # object from the 'connect'.
    
    if (not $self->{mcount}++) {
        $self->{msg} = delete $self->{dummy};
    }
    else {
        $self->{msg} = PureMessage::Policy::MessageObject->new(
            minfo => PureMessage::MessageInfo->new(%{$self->{con}}),
            logger => $self,
            quarantine => $quarantine,
        );
    }

    # Set limits on the message. These will be applied when parsing the
    # message from inside the Policy::Engine's process() method.
    $self->{msg}->set_limit(
	max_total_parts  => $config{mime_max_total_parts},
	max_part_depth   => $config{mime_max_part_depth},
	max_part_headers => $config{mime_max_part_headers},
    );

    $self->{msg}->set_attr('macro_daemon_name', $self->m_daemon_name);

    $self->{msg}->set_attr('envfrom_args', join(' ', @esmtp_args));
    $self->{msg}->minfo->value('Envelope-From', _addr_clean($sender));
    $self->{msg}->minfo_value('Queue-Id', $self->m_mail_queue_id);
    $self->{msg}->tmp_id($self->m_mail_queue_id);
    my @time = split(' ', gmtime);
    $self->{msg}->minfo_value('Date', sprintf("%s, %02d %s %4d %s GMT", @time[0,2,1,4,3]));
    $self->{recip} = [];
    return SMFIS_CONTINUE;
}

sub pmx_envrcpt {
    my ($self, $recipient, @esmtp_args) = @_;
    $self->log(LOG_DEBUG, "envrcpt: $recipient");
    push @{$self->{recip}}, $recipient;
    return SMFIS_CONTINUE;
}

sub pmx_eoh {
    my $self = shift;
    my $smtp_mode = $self->smtp_mode;
    my @recips = @{$self->{recip}};
    my $msg = $self->{msg};

    $_ = _addr_clean($_) foreach(@recips);
    $msg->minfo->value('Envelope-To', @recips);
    $msg->set_attr('envrcpt_count', scalar(@recips));

    my $header = "";
    if ($_[0]) {
	my $insert_rec_before = -1;
	if (!$smtp_mode ) {
	    # insert a faked-up received header - after the Return-Path header,
	    # if it is first, otherwise make it the first header.
	    if (lc $_[0][0] eq 'return-path') {
		$insert_rec_before = 2; # headers array counts up by 2
	    }
	    else {
		$insert_rec_before = 0;
	    }
	}
	my $rcount = 0;
	my $rfound;
	my $hdr_size = @{$_[0]};
        for (my $i = 0; $i < $hdr_size; $i += 2) {
	    
	    if ($i == $insert_rec_before) {
		my $relay = $msg->minfo_value('Relay');
		my $sm_host = $msg->minfo_value('Sendmail-Host');
		my $qid = $msg->minfo_value('Queue-Id');
		my @env_to = $msg->minfo_value('Envelope-To');
		my $env_from = $msg->minfo_value('Envelope-From');
		my $date = $msg->minfo_value('Date');
		my $helo = $msg->minfo_value('Helo');
		$header .=
		    "Received: from $helo ($relay)\n\tby $sm_host with SMTP id $qid";
		$header .= @env_to > 1 ? ";\r\n\t" : "\r\n\tfor $env_to[0]; ";
		$header .= "$date\r\n";
		$header .= "\t(envelope-from $env_from)\r\n";
	    }
	    
	    # Strip trailing vertical whitespace before building up the header
	    # string. This will never occur with Sendmail, but can occur with
	    # the less strict asmilter library.
	    $_[0][$i+1] =~ s/[\r\n]*$//;
	    
            $header .= $_[0][$i] . ': ' . $_[0][$i+1] . "\r\n";
            if (lc($_[0][$i]) eq 'x-pmx-quarantine-approved'
                and $_[0][$i+1] eq $config{quarantine_secret})
	    {
                $self->{approved} = 1;
                $self->log(LOG_INFO, "quarantine secret found");
                last;
            }
	    
	    # try to extract relay from Received header if in SMTP mode or the
	    # relay_use_header option is set
            my $relay;
	    if ( $relay_use_header
		and !$rfound
		and lc($_[0][$i]) eq lc($relay_header)
		and $rcount++ == $relay_index
		and $relay = _get_hdr_relay($_[0][$i+1]) ) {

                my $fields = _get_hdr_fields($_[0][$i+1]);
                if ($relay eq 'USE_MAIL_API') {
                    $relay = "$fields->{rdns} [$fields->{ip}]";
                }
                $msg->minfo_value('relay_auth', 1) if $fields->{auth};

		if (defined $relay and length $relay) {
		    $rfound = 1;
		    $msg->minfo_value('Relay', $relay);
		    # try to extract just the hostname for the message_log
		    $relay =~ s/^\s*(\S*).*$/$1/;
		    $relay =~ s/^\[(.*)\]$/$1/;
		    $self->mlog_relay(length $relay ? $relay : "(UNKNOWN)");
		}
		else {
		    my $val = $_[0][$i+1];
		    $msg->log(LOG_WARNING, "Warning: unable to extract relay data from \u$relay_header header (index=$relay_index, regex='$config{relay_from_regex}', value='$val')");
		    $self->mlog_relay('(UNKNOWN)');
		}
	    }
        }
	if ( $relay_use_header  and $rcount <= $relay_index) {
	    $msg->log(LOG_WARNING, "Warning: message contained only $rcount \u$relay_header headers, not enough to extract relay data (index=$relay_index)");
	    $self->mlog_relay('(UNKNOWN)');
	}
        if ($self->{approved}) {
            # we have to wait for eom so that
            # we can strip the approved header
            return SMFIS_CONTINUE;
        }
        $header .= "\r\n";
        $msg->print($header);

        eval {
            require PureMessage::Policy::BATV;
            require PureMessage::Policy::Spam::Config;

            my $strip = $PureMessage::Policy::Spam::Config::config{plugin}{batv}{strip};
            $strip = 1 unless defined $strip;
            PureMessage::Policy::BATV::pmx_batv_strip($msg) if $strip;
        };
    }
        
    # wait for eom
    if ($splitting) {
	# wait for eom
	return SMFIS_CONTINUE;
    }
    else {
	# set mapped_recipients on the message
	$msg->mapped_recipients($engine->map_recipients($msg, $msg->envelope_recipients));
    }
    
    # process message
    $msg->event('eoh');
    my $new = ($self->{mcount} != 1 or $smtp_mode);
    return $self->_process($msg, new => $new);
}

sub pmx_body {
    my $self = shift;
    $self->{msg}->print($_[0]) unless $self->{approved};
    return SMFIS_CONTINUE;
}

sub pmx_eom {
    my ($self) = @_;
    
    # approve queued message
    if (delete $self->{approved}) {
        # strip the approved header
        $self->chgheader("X-PMX-Quarantine-Approved", 1, "");
        $self->log(LOG_NOTICE, "accepted");
	_msg_cleanup($self);
        return SMFIS_ACCEPT;
    }
    
    my $main_msg = $self->{msg};
    $main_msg->event('eom');

    die "No recipients found - missed eoh callback?" unless defined $main_msg->envelope_recipients;

    # Determine first untrusted relay from headers
    my $fur_parts = $self->_get_fur($engine->get_relay_matchers());
    my $fur = 'none';
    if ($fur_parts) {
        $fur = $fur_parts->{ip} if $fur_parts->{ip};
        $main_msg->minfo->value('Fur_rdns', $fur_parts->{rdns})
            if $fur_parts->{rdns};
        $main_msg->minfo->value('Fur_helo', $fur_parts->{helo})
            if $fur_parts->{helo};
        $main_msg->minfo->value('Fur_auth', $fur_parts->{auth})
            if $fur_parts->{auth};
    }
    $main_msg->minfo->value('Fur', $fur);
    $main_msg->set_attr('mark_fur', $fur, 0);

    my %paths;
    if ($splitting && (%paths = $engine->get_paths($main_msg))) {
        # get all possible paths through the per-recipient tests of the script
        # for this message
        my @path_keys = keys %paths;

        # the existing message will use the first path, and will be
        # processed/delivered by the milter as normal
        my $first_path = shift @path_keys;
        my @first_path_env_to = @{$paths{$first_path}[0]};
        my @first_path_mapped = @{$paths{$first_path}[1]};
        $main_msg->envelope_recipients(@first_path_env_to);
        $main_msg->mapped_recipients(@first_path_mapped);

        # create additional message objects
        for (my $i = 0; $i < @path_keys; $i++) {
            # split off one message per path
            my $path = $path_keys[$i];
            my $recips = $paths{$path};
            my $msg = $main_msg->clone('Envelope-To', $recips->[0], 'Mapped-Recipients', $recips->[1]);
            # propagate key main message attributes, add to this list any 
            # attributes that are needed in the split copy as well as the master
            foreach my $attr (qw(envrcpt_count)) {
                my $val = $main_msg->get_attr($attr); 
                $msg->set_attr($attr, $val, 0) if $val;
            }
            $self->_process_split_msg($msg, $path);

            $self->_add_pmx_reason($main_msg,$msg->get_attr('action'),$msg);

            # get splitted msg's mark and put into main msg
            for my $attr ($msg->attrs) {
                next unless $attr =~ /^x[1m]mark_/;
                my $mval = $main_msg->get_attr($attr);
                my $sval = $msg->get_attr($attr);

                # handle multiple marks
                if ($attr =~ /^xmmark_/) {
                    if (defined $mval) {
                        push @{$mval}, @{$sval};
                    }
                    else {
                        $main_msg->set_attr($attr, $sval);
                    }
                }
                # handle single marks
                else {
                    $main_msg->set_attr($attr, $sval);
                }
            }

            $msg->free;
	}
	
	$self->log(LOG_DEBUG, "POLICY MILTER: processing message for '".
		    ( join ', ', @first_path_env_to ). "' (final - result will be returned to MTA)");
        
        # process - single message / first path
	return $self->_process($main_msg, path => $first_path, new => 1);
    }
    else {
	return $self->_process($main_msg);
    }
}

sub pmx_abort {
    my ($self) = @_;
    $self->log(LOG_NOTICE, "abort");
    _msg_cleanup($self);
    return SMFIS_CONTINUE;
}

sub pmx_close {
    my ($self) = @_;
    if (my $watch = delete $self->{watch}) {
	$self->log(LOG_INFO, "conn times: " . ActiveState::StopWatch::stop_watch($watch));
    }
    
    # clean up
    $self->closelog;
    my $dummy = delete $self->{dummy};
    $dummy->free if $dummy;
    delete $self->{mcount};
    delete $self->{con};
    # cleanup the message if it wasn't done already (e.g. if plugin died)
    _msg_cleanup($self) if $self->{msg};

    # did we clean all we expected
    if ($^W) {
	for my $k (sort keys %$self) {
	    next if $k =~ /^[A-Z]+$/;
	    next if $k =~ /^__p(erl)?mx/;
	    next if $k =~ /^_logger/;
	    warn "Attribute '$k' left in milter object";
	}
    }
    return SMFIS_CONTINUE;
}

sub _process_split_msg {
    my $self = shift;
    my $msg = shift;
    my $path = shift;
    
    my $recip = join ', ', $msg->minfo_value('Envelope-To');
    
    $self->log(LOG_DEBUG, "POLICY MILTER: processing message for '$recip'");
    
    $engine->process($msg, path=>$path, new=>1);
    
    $self->log(LOG_ERR, "POLICY MILTER: error: expected the engine to have stopped") unless $msg->stopped;
    
    my $da = $msg->delivery_action;
    my $action = $da->{action} || 'keep';
    my $reason = $da->{reason};
    
    # process actions - anything but 'quarantine' or 'keep' just goes away
    if ($action eq 'discard') {
        $self->log(LOG_NOTICE, "POLICY MILTER: copy for '$recip' discarded");
    }
    elsif ($action eq 'reject') {
        $self->log(LOG_NOTICE, "POLICY MILTER: copy for '$recip' rejected: $reason");
    }
    elsif ($action eq 'tempfail') {
        $self->log(LOG_NOTICE, "POLICY MILTER: copy for '$recip' tempfail: copy discarded");
    }
    elsif ($action eq 'quarantine') {
        $msg->file($reason, discard=>1);
        $self->log(LOG_NOTICE, "POLICY MILTER: copy for '$recip' quarantined: $reason");
    }
    elsif ($action eq 'route') {
        $self->log(LOG_NOTICE, "POLICY MILTER: copy for '$recip' routed: copy discarded");
    }
    else {
        $self->log(LOG_ERR, "POLICY MILTER: unknown delivery action '$action'") unless ($action eq 'keep');
        $reason = 'Queue';
        $msg->file($reason, discard=>1);
        $self->log(LOG_NOTICE, "POLICY MILTER: copy for '$recip' queued for delivery");
    }

    $msg->set_attr('action',$action);

}

sub _process {
    my $self = shift;
    my $msg = shift || $self->{msg};
    $engine->process($msg, @_);
    my $rc = SMFIS_CONTINUE;
    if ($msg->stopped) {
        my $da = $msg->delivery_action;
        my $action = $da->{action} || 'keep';
        my $reason = $da->{reason} || '';
    	my $event = $msg->event;

        $msg->fixup_message_id();
        if ($action eq 'discard') {
            $self->log(LOG_NOTICE, "discarded");
            $rc = SMFIS_DISCARD;
        }
        elsif ($action eq 'reject') {
            $self->log(LOG_NOTICE, "rejected");
            $self->setreply($da->{rcode}, $da->{xcode}, $reason);
            $rc = SMFIS_REJECT;
        }
        elsif ($action eq 'tempfail') {
            $self->log(LOG_NOTICE, "tempfail");
            $self->setreply($da->{rcode}, $da->{xcode}, $reason);
            $rc = SMFIS_TEMPFAIL;
        }
        elsif ($action eq 'keep')  {
            if (!$msg->modified or $event eq 'eom') {
                $rc = SMFIS_ACCEPT;
                $self->log(LOG_NOTICE, "accepted");
            }
        }
        elsif ($action eq 'route') {
            $self->log(LOG_NOTICE, "route: $reason");
            $rc = SMFIS_DISCARD;
        }
        elsif ($event eq 'eom') {
            if ($action eq 'quarantine') {
                $msg->file($reason, discard=>1);
                $self->log(LOG_NOTICE, "quarantine: $reason");
                $rc = SMFIS_DISCARD;
            }
            else {
                $self->log(LOG_ERR, "unknown delivery action '$action'");
                $self->log(LOG_NOTICE, "accepted");
                $rc = SMFIS_ACCEPT;
            }
        }

	if ($event eq 'eom' && $rc != SMFIS_DISCARD) {
	    # pick up changes to envelope recipients
	    if ($msg->rcpt_modified) {
		my @new = $msg->minfo_value('Envelope-To');
		my @orig = @{$self->{recip}};
                        
		# additions
		foreach my $recip (@new) {
		    $self->addrcpt($recip) unless 
                        grep lc($recip) eq lc(_addr_clean($_)), @orig;
		}
            
		# deletions
		foreach my $recip (@orig) {
		    $self->delrcpt($recip) unless 
                        grep lc(_addr_clean($recip)) eq lc($_), @new;
		}
	    }
            if ($msg->from_modified) {
		my $new = $msg->envelope_from();
		my $args = $msg->get_attr('envfrom_args');

		$self->chgfrom($new, $args);
            }

            # dkim should be done after all the changes made to header and body
            $msg->dkim_sign() if $msg->get_attr('dkim_sign');

	    # pick up changes to headers
	    if ($msg->header_modified) {
		my @mods = $msg->get_header_additions;
		for (my $i = 0; $i < scalar @mods; $i += 2) {
		    $self->addheader(@mods[$i,$i+1]);
		}
		@mods = $msg->get_header_replacements;
		for (my $i = 0; $i < scalar @mods; $i += 3) {
		    my @found_mods = @mods[$i..$i+2];
		    $found_mods[1]++; # milter indices are 1-based.
		    $self->chgheader(@found_mods);
		}            
	    }
	    # pick up changes to body
	    if ($msg->body_modified) {
		$msg->implode($self->can("replacebody"),
			      cb_param=>$self, no_header=>1);
	    }
	}

        $self->_add_pmx_reason($msg,$action);

        if ($rc != SMFIS_CONTINUE || $event eq 'eom') {
            # get mark attributes that may have been added to the message
            foreach my $attr ($msg->attrs) {
                next unless $attr =~ /^(?:x[1m])?(mark_.*)/;
                my $tattr = $1;
                my $value = $msg->get_attr($attr);
                my @value_ary = ref($value) eq 'ARRAY' ? @$value : ($value);
                foreach my $value (@value_ary) {
                    $self->mark(substr($tattr, 5), $value);
                }
            }
            _msg_cleanup($self);
        }
    }

    $rc;
}

sub _add_pmx_reason {
    my ($self,$primary_msg,$action,$cloned_msg) = @_;
   
    # if we don't have a cloned object, use the main msg object to get the
    # data we need for our pmx_action line 
    $cloned_msg = $primary_msg unless $cloned_msg;
         
    my $existing_value = $primary_msg->get_attr('mark_pmx_action');
    $existing_value = ref($existing_value) ? $existing_value : [];

    my $final_action = $action || '-';
    my $final_reason = $cloned_msg->get_attr("mark_pmx_reason") || '-';

    # resulting message log format is:
    #  pmx_action=ACTION,REASON,GROUP_MEMBERSHIP,ENV_RECIPIENT,MAPPED_RECIPIENT,..
    foreach my $this_recip ( $cloned_msg->minfo_value('Envelope-To') ) {                                          
        my @mrecip = $engine->map_recipients($cloned_msg,$this_recip);              
        my $group;
        if ($engine->{_pmxpe_gmap}) {
            my $direction = $engine->get_direction($primary_msg);
            if ($direction eq "Inbound") {
                $group = $engine->{_pmxpe_gmap}->lookup($mrecip[0]);
            }
            else {  
                my $minfo = $primary_msg->minfo;
                my $sender = $minfo->value('Envelope-From');
                $group = $engine->{_pmxpe_gmap}->lookup($sender);
            }
        }
        $group ||= '-';
        push(@$existing_value,join(',',"$final_action,$final_reason,$group,$this_recip",@mrecip));
    }
    $primary_msg->set_attr('mark_pmx_action',$existing_value);
}

sub _msg_cleanup {
    my $self = shift;
    delete $self->{paths};
    delete $self->{path_keys};   
    if (my $msg = delete $self->{msg}) {
	$msg->free;
    }
    delete $self->{recip};
    if (my $watch = delete $self->{msg_watch}) {
	$self->log(LOG_INFO, "msg times: " . ActiveState::StopWatch::stop_watch($watch));
    }
}

sub _addr_clean {
    my $addr = shift;
    $addr = substr($addr, 1, -1) if $addr =~ /^<.*>\z/;
    $addr;
}

sub _get_fur {
   my ($self, $matchers) = @_;
   my $msg = $self->{msg};

   my $RELAY_ADDR   = qr/^(.*)\s?\[(.+)\]$/;
   my $RCVD_ANTE_IP = qr/[\[( \t]/;
   my $RCVD_POST_IP = qr/[\]) \t;]/;
   my $IP_ADDR      = qr/(?:\d{1,3}\.){3}\d{1,3}/;

   my $relay = $msg->minfo_value("relay");
   my $rdns;

   if($relay =~ /$RELAY_ADDR/) {
       $rdns = $1;
       $relay = $2;
   }
   else {
       $relay = $UNKNOWN_RELAY_IP;
   }
   if (! $relay_use_header) {
       unless (keys(%$matchers) and _match_ip($matchers, $relay)) {
           my $helo = $msg->minfo_value('Helo');
           my $auth = $msg->minfo_value('relay_auth');
           return {ip => $relay, rdns => $rdns, helo => $helo, auth => $auth};
       }
   }

   my @headers;
   eval { @headers = $msg->header('received');};
   my %seen = ();
   for my $h (@headers) {
       while ($h =~ /$RCVD_ANTE_IP($IP_ADDR)$RCVD_POST_IP/g) {
           my $ip = $1;
           next if $seen{$ip}++;
           unless (_match_ip($matchers, $ip)) {
               my $fields = _get_hdr_fields($h);
               $fields->{ip} = $ip;
               return $fields;
           }
       }
   }
   return undef;
}

sub _match_ip {
    my ($matchers, $ip) = @_;

    foreach my $match (keys %$matchers) {
        return 1 if ($matchers->{$match}->($ip));
    }
    return 0;
}

sub _get_hdr_relay {
    my ($header) = @_;

    if ($config{relay_from_regex} ne $def_relay_regex) {
        if ($header =~ $relay_regex) {
            return "$1";
        }
        return undef;
    }
    else { # Use the Mail API
        return 'USE_MAIL_API';
    }
}

sub _get_hdr_fields {
    my ($header) = @_;
    my ($dom, $addr, $helo, $auth);

    my $IP_REGEX = qr/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
    my $received = Mail::Field->new('Received', $header);

    $dom  = 'unknown';
    $addr = $UNKNOWN_RELAY_IP;
    $helo = 'unknown';
    $auth = 0;
    if ($received) {
        my $results = $received->parse_tree();

        if (exists $results->{'from'}->{'domain'} && 
            $results->{'from'}->{'domain'} &&
            $results->{'from'}->{'domain'} !~ $IP_REGEX) {
            $dom = $results->{'from'}->{'domain'};
        }
        # see DEF47316: RDNS not set... 
        elsif ($config{use_helo_as_rdns}) { 
            if (exists $results->{'from'}->{'HELO'} &&
                $results->{'from'}->{'HELO'}) {
                if ($header =~ m/from\s.+\s\(\[.+?\]\)/) {
                   $dom = $results->{'from'}->{'HELO'}; 
                }
            }
        }

        if (exists $results->{'from'}->{'address'} &&
            $results->{'from'}->{'address'}) {
            my $tmp_addr = $results->{'from'}->{'address'};
            $tmp_addr =~ s/[\[\]]//g;

            # Ensure that the IP is surrounded by [] chars
            if (exists $results->{'from'}->{'whole'} &&
                $results->{'from'}->{'whole'} =~ qr{\[$tmp_addr\]} ) {
                $addr = $tmp_addr;
            }
        }

        if (exists $results->{'from'}->{'HELO'} &&
            $results->{'from'}->{'HELO'}) {
            $helo = $results->{'from'}->{'HELO'};
        }

        if (exists $results->{'comments'} &&
            ref($results->{'comments'}) eq 'ARRAY') {
            for my $comment (@{$results->{'comments'}}) {
                $auth = 1 if $comment =~ m/^\(Authenticated sender:.*\)/;
            }
        }
    }
    return {rdns => $dom, ip => $addr, helo => $helo, auth => $auth};
}

1;

__END__

=head1 NAME

PureMessage::Milter::Policy - Milter driver for the PureMessage Policy Engine

=head1 SYNOPSIS

 <milter Policy>
     filter = PureMessage::Milter::Policy
     message_log = message_log
 </milter>

=head1 DESCRIPTION

The C<PureMessage::Milter::Policy> is a filter module for use with the
PureMessage milter.  It makes it possible to run the PureMessage
Policy Engine as a milter service.

The C<PureMessage::Milter::Policy> is enabled by adding it as
C<filter> to a milter section in the F<pmx.conf> configuration file.
Conventionally it is contained in a milter section called C<Policy>.
The C<message_log> option must be enabled for the C<pmx_mark>
policy action to be effective.

The C<PureMessage::Milter::Policy> will run the PureMessage Policy
Engine from the F<etc/policy.siv> script.

=head1 COPYRIGHT

Copyright (c) 2003 ActiveState Corp.  All rights reserved.

=head1 SEE ALSO

L<pmx-policy>, L<pmx-milter>, L<pmx-config>

=cut
